/* 
	ctrlproxy: A modular IRC proxy
	ip: support for connecting and listening on ipv4 and ipv6
	(c) 2002 Jelmer Vernooij <jelmer@nl.linux.org>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define _GNU_SOURCE
#include "ctrlproxy.h"
#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <errno.h>
#include <netdb.h>
#include <string.h>
#include <fcntl.h>
#include <stdlib.h>

#undef G_LOG_DOMAIN
#define G_LOG_DOMAIN "socket"

static gboolean handle_in (GIOChannel *c, GIOCondition o, gpointer data);
static gboolean handle_disc (GIOChannel *c, GIOCondition o, gpointer data);

pid_t piped_child(char* const command[], int *f_in)
{
	pid_t pid;
	int sock[2];

	if(socketpair(PF_UNIX, SOCK_STREAM, AF_LOCAL, sock) == -1) {
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "socketpair: %s", strerror(errno));
		return -1;
	}

	*f_in = sock[0];

	fcntl(sock[0], F_SETFL, O_NONBLOCK);

	pid = fork();
	if (pid == -1) {
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "fork: %s", strerror(errno));
		return -1;
	}

	if (pid == 0) {
		close(0);
		close(1);
		close(2);
		close(sock[0]);

		dup2(sock[1], 0);
		dup2(sock[1], 1);
		execvp(command[0], command);
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Failed to exec %s : %s\n", command[0], strerror(errno));
		return -1;
	}

	close(sock[1]);

	return pid;
}

void socket_to_iochannel(int sock, struct transport_context *c)
{
	GIOChannel *ioc;
	GError *error = NULL;
	ioc = g_io_channel_unix_new(sock);
	
	g_io_channel_set_encoding(ioc, NULL, &error);
	if(error)g_error_free(error);

	g_io_add_watch(ioc, G_IO_IN, handle_in, (gpointer)c);
	g_io_add_watch(ioc, G_IO_HUP, handle_disc, (gpointer)c);
	
	c->data = ioc;
}

int connect_pipe(struct transport_context *c) 
{
	char *args[100];
	int i;
	int argc = 0;
	int sock;
	pid_t pid;
	xmlNodePtr cur = c->configuration->xmlChildrenNode;
	memset(args, 0, sizeof(char *) * 100);


	while(cur) {
		if(xmlIsBlankNode(cur) || !strcmp(cur->name, "comment")) { 
			cur = cur->next; 
			continue; 
		}

		if(!strcmp(cur->name, "path")) args[0] = xmlNodeGetContent(cur);
		else if(!strcmp(cur->name, "arg")) args[++argc] = xmlNodeGetContent(cur);
		else g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Unknown element %s\n", cur->name);
	
		cur = cur->next;
	}

	xmlSetProp(c->configuration, "name", args[0]);
	
	args[++argc] = NULL;
	
	pid = piped_child(args, &sock);
	
	for(i = 0; i < argc; i++) xmlFree(args[i]);
	
	if(pid == -1) return -1;

	socket_to_iochannel(sock, c);
	
	return 0;
}

static gboolean handle_new_client(GIOChannel *c, GIOCondition o, gpointer data)
{
	GIOChannel *c_new;
	struct sockaddr clientname;
	struct transport_context *newcontext, *oldcontext = (struct transport_context *)data;
	GError *error = NULL;
	size_t size;
	int sock;

	size = sizeof(clientname);
	sock = accept(g_io_channel_unix_get_fd(c), (struct sockaddr *)&clientname, &size);
	if(!sock) {
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Can't accept connection!");
		return FALSE;
	}

	g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "New client %d", sock);

	c_new = g_io_channel_unix_new(sock);

	g_io_channel_set_encoding(c_new, NULL, &error);
	if(error)g_error_free(error);

	newcontext = malloc(sizeof(struct transport_context));
	memset(newcontext, 0, sizeof(struct transport_context));
	
	g_io_add_watch(c_new, G_IO_IN, handle_in, (gpointer)newcontext);
	g_io_add_watch(c_new, G_IO_HUP, handle_disc, (gpointer)newcontext);
	
	newcontext->data = c_new;
	newcontext->functions = oldcontext->functions;

	if(oldcontext->on_new_client) oldcontext->on_new_client(oldcontext, newcontext, oldcontext->caller_data);
}

static gboolean handle_disc (GIOChannel *ioc, GIOCondition o, gpointer data)
{
	struct transport_context *c = (struct transport_context *)data;

	if(c->on_disconnect)c->on_disconnect(c, c->caller_data);

	return TRUE;
}

static gboolean handle_in (GIOChannel *ioc, GIOCondition o, gpointer data)
{	
	struct transport_context *c = (struct transport_context *)data;
	GIOStatus status;
	char *ret;
	GError *error = NULL;

	g_assert(o == G_IO_IN);

	status = g_io_channel_read_line(ioc, &ret, NULL, NULL, &error);

	switch(status){
	case G_IO_STATUS_ERROR:
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "%s", (error?error->message:""));
		if(error)g_error_free(error);
		return TRUE;

	case G_IO_STATUS_EOF:
		if(error)g_error_free(error);
		if(c->on_disconnect)c->on_disconnect(c, c->caller_data);
		return FALSE;

	case G_IO_STATUS_AGAIN:
		if(error)g_error_free(error);
		return TRUE;

	case G_IO_STATUS_NORMAL: 
		if(c->on_receive)c->on_receive(c, ret, c->caller_data);
		free(ret);
		return TRUE;

	default: break; 
	}
	g_assert(0); 
	return TRUE;
}

int connect_ipv4(struct transport_context *c) 
{
	GIOChannel *ioc;
	struct sockaddr_in name;
	char *hostname = xmlGetProp(c->configuration, "host");
	char *aPort = xmlGetProp(c->configuration, "port");
	uint16_t port;
	struct hostent *hostinfo;
	int sock;
	char *tempname;
	GError *error = NULL;

	port = aPort?atoi(aPort):6667;
	if(aPort)xmlFree(aPort);

	g_assert(hostname);
	
	asprintf(&tempname, "%s:%d", hostname, port);

	xmlSetProp(c->configuration, "name", tempname);

	g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Connecting to %s:%d", hostname, port);

	sock = socket (PF_INET, SOCK_STREAM, 0);
	if (sock < 0)
	{
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "(While connecting to server %s with IPv4): socket: %s", tempname, strerror(errno));
		free(tempname);
		return -1;
	}

	name.sin_family = AF_INET;
	name.sin_port = htons (port);
	hostinfo = gethostbyname (hostname);
	if (hostinfo == NULL)
	{
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_ERROR, "Unknown host %s.\n", hostname);
		xmlFree(hostname);
		free(tempname);
		return -1;
	}

	name.sin_addr = *(struct in_addr *) hostinfo->h_addr;
	if (0 > connect (sock,
					 (struct sockaddr *) &name,
					 sizeof (name)))
	{
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "(While connecting to server %s): connect: %s", tempname, strerror(errno));
		free(tempname);
		return -1;
	}
	free(tempname);

	socket_to_iochannel(sock, c);

	g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Successfully connected to %s:%d", hostname, port);
	xmlFree(hostname);
	return 0;
}



int connect_ipv6(struct transport_context *c)
{
	GIOChannel *ioc;
	struct sockaddr_in6 name;
	char *hostname = xmlGetProp(c->configuration, "host");
	char *aPort = xmlGetProp(c->configuration, "port");
	uint16_t port;
	struct hostent *hostinfo;
	int sock;
	char *tempname;
	GError *error = NULL;

	port = aPort?atoi(aPort):6667;
	if(aPort)xmlFree(aPort);

	g_assert(hostname);
	
	asprintf(&tempname, "%s:%d", hostname, port);

	xmlSetProp(c->configuration, "name", tempname);

	g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Connecting to %s:%d", hostname, port);

	sock = socket (PF_INET6, SOCK_STREAM, 0);
	if (sock < 0)
	{
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "(While connecting to server %s with IPv6): socket: %s", tempname, strerror(errno));
		free(tempname);
		return -1;
	}

	name.sin6_family = AF_INET6;
	name.sin6_port = htons (port);
	hostinfo = gethostbyname2 (hostname, AF_INET6);
	if (hostinfo == NULL)
	{
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_ERROR, "Unknown host %s.\n", hostname);
		xmlFree(hostname);
		free(tempname);
		return -1;
	}
	
	memcpy((char *)&name.sin6_addr, hostinfo->h_addr, hostinfo->h_length);
	if (0 > connect (sock,
					 (struct sockaddr *) &name,
					 sizeof (name)))
	{
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "(While connecting to server %s with IPv6): connect: %s", tempname, strerror(errno));
		free(tempname);
		return -1;
	}
	free(tempname);

	socket_to_iochannel(sock, c);

	g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Successfully connected to %s:%d", hostname, port);
	xmlFree(hostname);
	return 0;
}

int make_socket (uint16_t port)
{
	int sock;
	struct sockaddr_in name;

	/* Create the socket. */
	sock = socket (PF_INET, SOCK_STREAM, 0);
	if (sock < 0)
	{
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "socket: %s\n", strerror(errno));
		return -1;
	}

	/* Give the socket a name. */
	name.sin_family = AF_INET;
	name.sin_port = htons (port);
	name.sin_addr.s_addr = htonl (INADDR_ANY);
	if (bind (sock, (struct sockaddr *) &name, sizeof (name)) < 0)
	{
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "bind: %s\n", strerror(errno));
		return -1;
	}

	return sock;
}

int close_socket(struct transport_context *c)
{
	GIOChannel *gio;
	GError *error = NULL;

	gio = c->data;

	g_io_channel_shutdown(gio, TRUE, &error);
	if(error)g_error_free(error);

	g_io_channel_unref(gio);

	c->data = NULL;
}

int write_socket(struct transport_context *t, char *l)
{
	GError *error = NULL;
	GIOChannel *c = t->data;

	if(!c) {
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "Trying to send line '%s' to socket that is not connected!\n", l);
		return -1;
	}
	
	if(g_io_channel_write_chars(c, l, -1, NULL, &error) == G_IO_STATUS_ERROR)
	{ 
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Can't send: %s", (error?error->message:"g_io_channel_write_chars failed"));
		g_error_free(error);
		return -1;
	}

	g_assert(!error);
	g_io_channel_flush(c, &error);

	if(error) {
		g_error_free(error);
		return -1;
	}

	return 0;
}

int listen_ip(struct transport_context *c)
{
	/* FIXME: Work for ipv6 as well */
	static int client_port = 6667;
	int port, sock;
	GIOChannel *gio;
	GError *error = NULL;
	
	if(xmlHasProp(c->configuration, "port")) {
		char *aport = xmlGetProp(c->configuration, "port");
		port = atoi(aport);
		xmlFree(aport);
	} else port = ++client_port;

	sock = make_socket(port);

	if(sock == -1) {
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Error trying to create socket for port %d: %s", port, strerror(errno));
		return -1;
	}

	if(listen(sock, 5) < 0) {
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Error trying to listen on port %d: %s", port, strerror(errno));
		return -1;
	}

	g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "Listening on port %d(socket %d)", port, sock);

	gio = g_io_channel_unix_new(sock);

	g_io_channel_set_encoding(gio, NULL, &error);
	if(error)g_error_free(error);

	g_io_add_watch(gio, G_IO_IN, handle_new_client, c);
	g_io_channel_unref(gio);

	c->data = gio;

	return 0;
}

struct transport ipv4 = {
	"ipv4",
	connect_ipv4,
	listen_ip,
	write_socket,
	close_socket
};

struct transport ipv6 = {
	"ipv6",
	connect_ipv6,
	listen_ip,
	write_socket,
	close_socket
};

struct transport pipe_transport = {
	"pipe",
	connect_pipe,
	NULL,
	write_socket,
	close_socket
};

void init_plugin(GModule *g, xmlNodePtr p) {
	register_transport(&ipv4);
	register_transport(&ipv6);
	register_transport(&pipe_transport);
}
