/* 
	ctrlproxy: A modular IRC proxy
	(c) 2002 Jelmer Vernooij <jelmer@nl.linux.org>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "internals.h"

static size_t get_new_id(void) 
{
	static size_t newid = 0;
	return (newid++);
}

static void free_names(struct channel *c) 
{
	struct nick *n;
	GList *g = c->nicks;
	while(g) {
		n = (struct nick *)g->data;
		free(n->name); free(n);
		g = g_list_next(g);
	}
	g_list_free(c->nicks);
	c->nicks = NULL;
}

static void free_channel(struct channel *c)
{
	free_names(c);
	if(c->topic)free(c->topic);
	c->topic = NULL;
}

struct channel *find_channel(struct network *st, char *name) 
{
	GList *cl = st->channels;
	while(cl) {
		struct channel *c = (struct channel *)cl->data;
		char *channel_name = xmlGetProp(c->xmlConf, "name");
		if(!strcasecmp(channel_name, name))return c;
		xmlFree(channel_name);
		cl = g_list_next(cl);
	}
	return NULL;
}

struct channel *find_add_channel(struct network *st, char *name) {
	struct channel *c = find_channel(st, name);
	if(c)return c;
	c = malloc(sizeof(struct channel));
	memset(c, 0, sizeof(struct channel));
	c->id = get_new_id();
	st->channels = g_list_append(st->channels, c);

	/* check if there is a XML node for this channel yet */
	c->xmlConf = xmlFindChildByName(st->xmlConf, name);
	
	if(!c->xmlConf) {
		c->xmlConf = xmlNewNode(NULL, "channel");
		xmlSetProp(c->xmlConf, "name", name);
		xmlAddChild(st->xmlConf, c->xmlConf);
	}
	
	return c;
}

struct nick *find_nick(struct channel *c, char *name) {
	GList *l = c->nicks;
	struct nick *n;
	char *realname = name;
	if(realname[0] == '@' || realname[0] == '+')realname++;

	while(l) {
		n = (struct nick *)l->data;
		if(!strcasecmp(n->name, realname))return n;
		l = g_list_next(l);
	}

	return NULL;
}

struct nick *find_add_nick(struct channel *c, char *name) {
	struct nick *n = find_nick(c, name);
	char *realname = name;
	if(n)return n;
	if(strlen(name) == 0)return NULL;
	n = malloc(sizeof(struct nick));
	memset(n, 0, sizeof(struct nick));
	if(realname[0] == '@' || realname[0] == '+'){
		n->mode = realname[0];
		realname++;
	}
	n->id = get_new_id();
	n->name = strdup(realname);
	c->nicks = g_list_append(c->nicks, n);
	return n;
}


static void handle_join(struct network *s, struct line *l) 
{
	struct channel *c;
	/* Someone else is joining a channel the user is on */
	if(l->direction == FROM_SERVER && line_get_nick(l)) {
		c = find_add_channel(s, l->args[1]);
		find_add_nick(c, line_get_nick(l));
	}
	/* The user is joining a channel */
	else if(l->direction == TO_SERVER) {
		c = find_add_channel(s, l->args[1]);
		xmlSetProp(c->xmlConf, "autojoin", "1");
	}
}


static void handle_part(struct network *s, struct line *l) 
{
	struct channel *c;
	struct nick *n;

	if(!line_get_nick(l))return;
	c = find_channel(s, l->args[1]);

	if(l->direction == TO_SERVER && c) {
		s->channels = g_list_remove(s->channels, c);
		free_channel(c);
		free(c);
		xmlSetProp(c->xmlConf, "autojoin", "0");
		return;
	} 
	
	if(c){
		n = find_nick(c, line_get_nick(l));
		if(n) {
			free(n->name);
			c->nicks = g_list_remove(c->nicks, n);
			free(n);
		} else g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Can't remove nick %s from channel %s: nick not on channel\n", line_get_nick(l), l->args[1]);

		return;
	} 
	
	g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Can't part or let other nick part %s(unknown channel)\n", l->args[1]);
}

static void handle_kick(struct network *s, struct line *l) {
	struct channel *c = find_channel(s, l->args[1]);
	if(c){
		struct nick *n = find_nick(c, l->args[2]);
		if(n) {
			free(n->name);
			c->nicks = g_list_remove(c->nicks, n);
			free(n);
		} else g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Can't kick nick %s from channel %s: nick not on channel\n", l->args[2], l->args[1]);
	} else g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Can't kick nick %s from %s\n", l->args[2], l->args[1]);
}

static void handle_topic(struct network *s, struct line *l) {
	struct channel *c = find_channel(s, l->args[1]);
	if(c->topic)free(c->topic);
	c->topic = strdup(l->args[2]);
}

static void handle_332(struct network *s, struct line *l) {
	struct channel *c = find_channel(s, l->args[2]);
	if(c->topic)free(c->topic);
	c->topic = strdup(l->args[3]);
}

static void handle_no_topic(struct network *s, struct line *l) {
	struct channel *c = find_channel(s, l->args[1]);
	if(c->topic)free(c->topic);
	c->topic = NULL;
}

static void handle_namreply(struct network *s, struct line *l) {
	char *names, *tmp, *t;
	struct channel *c = find_channel(s, l->args[3]);
	if(c) {
		c->mode = l->args[2][0];
		if(c->namreply_started == 0) {
			free_names(c);
			c->namreply_started = 1;
		}
		tmp = names = strdup(l->args[4]);
		while((t = strchr(tmp, ' '))) {
			*t = '\0';
			find_add_nick(c, tmp);
			tmp = t+1;
		}
		find_add_nick(c, tmp);
		free(names);
	} else g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Can't add names to %s: channel not found\n", l->args[3]);
}

static void handle_end_names(struct network *s, struct line *l) {
	struct channel *c = find_channel(s, l->args[2]);
	if(c)c->namreply_started = 0;
	else g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Can't end /NAMES command for %s: channel not found\n", l->args[2]);
}

static void handle_quit(struct network *s, struct line *l) {
	if(!line_get_nick(l))return;
	GList *g = s->channels;
	while(g) {
		struct channel *c = (struct channel *)g->data;
		struct nick *n = find_nick(c, line_get_nick(l));
		if(n) {
			free(n->name);
			c->nicks = g_list_remove(c->nicks, n);
			free(n);
		}
		g = g_list_next(g);
	}
}

static void handle_mode(struct network *s, struct line *l) 
{
	/* FIXME: Store stuff */
}

static void handle_nick(struct network *s, struct line *l)
{
	if(l->direction == FROM_SERVER && l->args[1] && line_get_nick(l)) {
		GList *g = s->channels;
		while(g) {
			struct channel *c = (struct channel *)g->data;
			struct nick *n = find_nick(c, line_get_nick(l));
			if(n) {
				free(n->name);
				n->name = strdup(l->args[1]);
			}
			g = g_list_next(g);
		}
	}
	
	/* FIXME: Also handle our own users' nick changes! */
}

void state_reconnect(struct network *s) 
{
	GList *l = s->channels;

	/* Remove list of channels */
	while(l) {
		struct channel *ch = (struct channel *)l->data;
		free_channel(ch);
		free(ch);
		l = g_list_next(l);
	}

	g_list_free(s->channels);

	s->channels = NULL;
}

static struct irc_command {
	char *command;
	int min_args;
	void (*handler) (struct network *s, struct line *l);
} irc_commands[] = {
	{ "JOIN", 1, handle_join },
	{ "PART", 1, handle_part },
	{ "KICK", 2, handle_kick },
	{ "QUIT", 0, handle_quit },
	{ "TOPIC", 2, handle_topic },
	{ "332",  3, handle_332 },
	{ "331" , 1, handle_no_topic },
	{ "353" , 4, handle_namreply },
	{ "366",  2, handle_end_names },
	{ "NICK" , 1, handle_nick },
	{ "MODE", 0, handle_mode },
	{ NULL }
};

void state_handle_data(struct network *s, struct line *l)
{
	int i,j;
	
	if(!l->args[0])return;

	for(i = 0; irc_commands[i].command; i++) {
		if(!strcasecmp(irc_commands[i].command, l->args[0])) {
			for(j = 0; j <= irc_commands[i].min_args; j++) {
				if(!l->args[j])return;
			}
			irc_commands[i].handler(s,l);
			return;
		}
	}
	return;
}


GSList *gen_replication(struct network *s)
{
	struct nick *n;
	GList *cl = s->channels, *nl;
	struct channel *c;
	GSList *ret = NULL;
	char *nick, *server_name, *channel_name;

	while(cl) {
		c = (struct channel *)cl->data;
		channel_name = xmlGetProp(c->xmlConf, "name");
		if(channel_name[0] != '#' && channel_name[0] != '&') {
			cl = g_list_next(cl);
			xmlFree(channel_name);
			continue;
		}
		nick = xmlGetProp(s->xmlConf, "nick");
		server_name = xmlGetProp(s->xmlConf, "name");
		ret = g_slist_append(ret, irc_parse_linef(":%s JOIN %s\r\n", nick, channel_name));

		if(c->topic) {
			ret = g_slist_append(ret, irc_parse_linef(":%s 332 %s %s :%s\r\n", server_name, nick, channel_name, c->topic));
		} else {
			ret = g_slist_append(ret, irc_parse_linef(":%s 331 %s %s :No topic set\r\n", server_name, nick, channel_name));
		}

		nl = c->nicks;
		while(nl) {
			n = (struct nick *)nl->data;
			if(n->mode) { ret = g_slist_append(ret, irc_parse_linef(":%s 353 %s %c %s :%c%s\r\n", server_name, nick, c->mode, channel_name, n->mode, n->name)); }
			else { ret = g_slist_append(ret, irc_parse_linef(":%s 353 %s %c %s :%s\r\n", server_name, nick, c->mode, channel_name, n->name)); }
			nl = g_list_next(nl);
		}
		ret = g_slist_append(ret, irc_parse_linef(":%s 366 %s %s :End of /names list\r\n", server_name, nick, channel_name));
		c->introduced = 3;
		cl = g_list_next(cl);
		xmlFree(channel_name);
		xmlFree(nick);
		xmlFree(server_name);
	}
	return ret;
}
